#include "squidwall.h"

/* filteroptions for content scanning
 *****************************************************
 *
 * 1. content filtering
 * +c,Filter1,Filter2,Filter3
 * -c disable content filtering
 * ---------------------------------------------------
 *
 * 2. content filtering, using regex.7 expressions
 * +C,Filter1,Filter2,Filter3
 * -C disable regex content filtering
 * ---------------------------------------------------
 *
 * 3. url filtering
 * +u,Filter1,Filter2,Filter3
 * -u disable url filtering
 * ---------------------------------------------------
 *
 * 4. url filtering, using regex.7 expressions
 * +U,Filter1,Filter2,Filter3
 * -U disable regex url filtering
 * ---------------------------------------------------
 *
 * 5. antivirus scanning
 * +V
 * -V disable antivirus scanning
 * ---------------------------------------------------
 */

#define FILTER_STRING 0
#define FILTER_REGEX  1

/* freeing allocated memory */
static void read_filters_freeregex(array *root);
static void read_filters_freestring(array *root);

/* read the acl's into memory */
static void read_filters_regex(array *a, char *start);
static void read_filters_string(array *a, char *start);
static void read_filters_file_regex(array *root, char *file);
static void read_filters_file_string(array *root, char *file);

static void read_filters_allfiles(array *a, int type);

/* current working directory */
char *filters_dir;

/* remember, what filters are read allready */
int todo_done;

/*****************************************************************/
static void read_filters_freeregex(array *root)
{
	struct squidwall_regex *a;
	int64 l;

	for (;;) {
		l=array_length(root, sizeof(struct squidwall_regex));
		if (l == 0) break;

		a = array_get(root, sizeof(struct squidwall_regex), l-1);
		if (!a) { break; } /* can't happen! */

		/* free allocated things */
		stralloc_free(&a->name);
		regfree(&a->regex);
	}
	/* free tho whole array */
	array_reset(root);

	return;
}

static void read_filters_freestring(array *root)
{
	struct squidwall_string *a;
	int64 l;

	for (;;) {
		l=array_length(root, sizeof(struct squidwall_string));
		if (l == 0) break;

		a = array_get(root, sizeof(struct squidwall_string), l-1);
		if (!a) { break; } /* can't happen! */

		/* free allocated things */
		stralloc_free(&a->name);
		stralloc_free(&a->string);
	}
	/* free tho whole array */
	array_reset(root);

	return;
}

/*****************************************************************/
static void read_filters_regex(array *a, char *start)
{
	char *x=start;
	int i=0, l=str_len(start);

	while (*++x) {

		/* read all available filters and return */
		if (*x == '*') {
			read_filters_allfiles(a, FILTER_REGEX);
			return;
		}

		/* ',' -> '\0' */
		if (*x == ',') *x = 0;
	}

	/* read and compile each given filter */
	for (x=start+1; i < l; i+=str_len(x+i)+1) {
		read_filters_file_regex(a, x+i);
	}

	return;
}

static void read_filters_string(array *a, char *start)
{
	char *x=start;
	int i=0, l=str_len(start);

	while (*++x) {

		/* read all available filters and return */
		if (*x == '*') {
			read_filters_allfiles(a, FILTER_STRING);
			return;
		}

		/* ',' -> '\0' */
		if (*x == ',') *x = 0;
	}

	/* read and compile each given filter */
	for (x=start+1; i < l; i+=str_len(x+i)+1) {
		read_filters_file_string(a, x+i);
	}

	return;
}

/*****************************************************************/
static void read_filters_file_regex(array *root, char *file)
{
	struct squidwall_regex *r;
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, filters_dir)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 */
	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;

		/* remove \r\n and \n */
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;

		/* add to regex filter */
		r=array_allocate(root, sizeof(struct squidwall_regex),
			array_length(root, sizeof(struct squidwall_regex)));
		if (!r) die_nomem();
		byte_zero(r, sizeof(struct squidwall_regex));
		if (regcomp(&r->regex, sa.s, SW_REGCOMP_CFLAGS) != 0) {
			panic("regcomp() failed!");
			continue; /* read next line */
		}

		/* save the filtername (is used for the error reporting website) */
		stralloc_init(&r->name);
		if (!stralloc_copys(&r->name, file)) die_nomem();
		if (!stralloc_cats(&r->name, " (")) die_nomem();
		if (!stralloc_cats(&r->name, sa.s)) die_nomem();
		if (!stralloc_cats(&r->name, ")")) die_nomem();
		r->name.s[r->name.len]=0;
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_filters_file_string(array *root, char *file)
{
	struct squidwall_string *s;
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, filters_dir)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 */
	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;

		/* remove \r\n and \n */
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;

		/* add string to filter */
		s=array_allocate(root, sizeof(struct squidwall_string),
			array_length(root, sizeof(struct squidwall_string)));

		if (!s) die_nomem();
		byte_zero(s, sizeof(struct squidwall_string));

		/* save the string to search for ... */
		stralloc_init(&s->string);
		if (!stralloc_copys(&s->string, sa.s)) die_nomem();
		s->string.s[s->string.len]=0;

		/* save the filtername (is used for the error reporting website) */
		stralloc_init(&s->name);
		if (!stralloc_copys(&s->name, file)) die_nomem();
		if (!stralloc_cats(&s->name, " (")) die_nomem();
		if (!stralloc_cats(&s->name, sa.s)) die_nomem();
		if (!stralloc_cats(&s->name, ")")) die_nomem();
		s->name.s[s->name.len]=0;
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

/*****************************************************************/
static void read_filters_allfiles(array *a, int type)
{
	DIR *D;
	struct dirent *d;

	D=opendir(filters_dir);
	if (!D) return;
	while ((d=readdir(D))) {

		/* skip '.' and '..' */
		if (d->d_name[0]=='.' && d->d_name[1]==0) continue;
		if (d->d_name[0]=='.' && d->d_name[1]=='.' && d->d_name[2]==0)
			continue;

		switch (type) {
			case FILTER_STRING:
				read_filters_file_string(a, d->d_name);
				break;
			case FILTER_REGEX:
				read_filters_file_regex(a, d->d_name);
				break;
		}
	}
	closedir(D);

	return;
}

void read_filters(struct squidwall_acl *acl, const char *path, const char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 e.g. */
	if (buffer_mmapread(&b, sa.s) == -1) {
		/* if the host/user failed... read the '!'
		 * - /etc/squidwall/hosts/!
		 * - /etc/squidwall/users/!
		 */
		if (!stralloc_copys(&sa, path)) die_nomem();
		if (!stralloc_cats(&sa, "/!")) die_nomem();
		sa.s[sa.len]=0;
		if (buffer_mmapread(&b, sa.s) == -1) return;
	}

	/* set default values for the acl */
	acl->todo=ACL_DEFAULT;

	/* nothing was read currently */
	todo_done=TODO_NONE;

	/* what is wanted */
	for (;;) {
		if (buffer_getnewline_sa(&b, &sa) != 1) break; /* 0=EOF -1=error 1=ok */
		sa.s[sa.len-1]=0; /* remove newline */

		if (sa.len < 2) continue; /* should not happen */
		switch (sa.s[0]) {
		case '+':
			switch (sa.s[1]) {
			case 'c':
				set(acl->todo, TODO_CONT);
				set(todo_done, TODO_CONT);
				filters_dir=cfilters_dir;
				read_filters_string(&acl->cfilter, sa.s+2);
				break;
			case 'C':
				set(acl->todo, TODO_CONTr);
				set(todo_done, TODO_CONTr);
				filters_dir=crfilters_dir;
				read_filters_regex(&acl->crfilter, sa.s+2);
				break;
			case 'u':
				set(acl->todo, TODO_URL);
				set(todo_done, TODO_URL);
				filters_dir=ufilters_dir;
				read_filters_string(&acl->ufilter, sa.s+2);
				break;
			case 'U':
				set(acl->todo, TODO_URLr);
				set(todo_done, TODO_URLr);
				filters_dir=urfilters_dir;
				read_filters_regex(&acl->urfilter, sa.s+2);
				break;
			case 'V':
				set(acl->todo, TODO_VIRUS);
				break;
			} /* switch sa.s[1] */
			break;
		case '-':
			switch (sa.s[1]) {
			case 'c':
				unset(acl->todo, TODO_CONT);
				set(todo_done, TODO_CONTr);
				filters_dir=cfilters_dir;
				read_filters_freestring(&acl->cfilter);
				break;
			case 'C':
				unset(acl->todo, TODO_CONTr);
				set(todo_done, TODO_CONTr);
				filters_dir=cfilters_dir;
				read_filters_freeregex(&acl->crfilter);
				break;
			case 'u':
				unset(acl->todo, TODO_URL);
				set(todo_done, TODO_URL);
				read_filters_freestring(&acl->ufilter);
				break;
			case 'U':
				unset(acl->todo, TODO_URLr);
				set(todo_done, TODO_URLr);
				read_filters_freeregex(&acl->urfilter);
				break;
			case 'V':
				unset(acl->todo, TODO_VIRUS);
				break;
			} /* switch sa.s[1] */
			break;
		} /* switch sa.s[0] */
	}

	/* read defaults, if no reading was done allready */
	if (!isset(todo_done, TODO_CONT)) {
		filters_dir=cfilters_dir;
		read_filters_string(&acl->cfilter, "*");
	}

	if (!isset(todo_done, TODO_CONTr)) {
		filters_dir=crfilters_dir;
		read_filters_regex(&acl->crfilter, "*");
	}

	if (!isset(todo_done, TODO_URL)) {
		filters_dir=ufilters_dir;
		read_filters_string(&acl->ufilter, "*");
	}

	if (!isset(todo_done, TODO_URLr)) {
		filters_dir=urfilters_dir;
		read_filters_regex(&acl->urfilter, "*");
	}

	buffer_close(&b);
	stralloc_free(&sa);

	return;
}
